#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include <stdlib.h>
#include "lcd.h"

#define F_CPU 1000000

/* Punctul va fi reprezentat printr-un bloc de 2 x 2 pixeli */
#define SIZE 2

/* Dimensiunea initiala a sarpelui */
#define INIT_LEN 4

/* Pozitia initiala a sarpelui */
#define INIT_X	30
#define INIT_Y	40

/* Dimensiunea maxima a sarpelui */
#define MAX_SNAKE_SIZE	100

/* Modurile de joc */
#define EASY	0
#define MEDIUM	1
#define HARD	2

/* Variabila ce retine modul de joc */
unsigned int play_mode;

/* Variabila ce retine scorul acumulat */
int scor = 0;

/* structura ce defineste un punct din joc avand SIZE * SIZE pixeli */
typedef struct {
	char x;
	char y;
} Point;

/* SNAKE */
Point snake[MAX_SNAKE_SIZE];
int lungime;

/* Punctul ce reprezinta mancarea */
Point food;

/* directie pentru snake */
#define SUS 1
#define STG 2
#define DRT 3
#define JOS 4
/* Directia actuala */
unsigned char directia;
/* Directia dupa apasarea unui buton */
unsigned char newDirection;

void timer_init(void) {

	sei();

	// Set Timer0 for generating a 10ms (2HZ) interrupt on channel A
	TCCR0A = (1 << WGM01);					// Enable CTC
	TCCR0B = (1 << CS02) | (1 << CS00);		// Set prescaler to 1:1024
	OCR0A = 0x9C;							// Set count to generate the signal
	TIMSK0 |= (1 << OCIE0A);				// Enable timer interrupts
}

ISR(TIMER0_COMPA_vect) {
}

/*
Deseneaza sarpele, mancarea si in functie de modul ales peretii si obstacolele */

void drawBlocks(void) {

	int i, j, k;
	char x1, x2, y1, y2;
	
	/* Deseneaza sarpele */
	for (i = 0; i < lungime; i++) {
	
		x1 = snake[i].x;
		y1 = snake[i].y;
		x2 = snake[i].x + SIZE;
		y2 = snake[i].y + SIZE;
		
		for(j = x1; j < x2; j++) {
			for(k = y1; k < y2; k++) {
				lcd_setpixel(j, k);
			}
		}
	}
	
	/* Deseneaza punctul de mancare */
	x1 = food.x;
	y1 = food.y;
	x2 = food.x + SIZE;
	y2 = food.y + SIZE;
	
	for(j = x1; j < x2; j++) {
		for(k = y1; k < y2; k++) {
			lcd_setpixel(j, k);
		}
	}
		
	/* Daca modul ales este HARD|MEDIUM desenam peretii */	
	if ((play_mode == MEDIUM) || (play_mode == HARD)) {
	
		for(i = 0; i < LCD_Y_RES; i++) {
			lcd_setpixel(0, i);
			lcd_setpixel(LCD_X_RES - 1, i);
		}
		
		for(i = 0; i < LCD_X_RES; i++) {
			lcd_setpixel(i, 0);
			lcd_setpixel(i, LCD_Y_RES - 1);
		}
	}
	
	/* Pentru modul hard adauga ostacole la pozitii standard */
	if (play_mode == HARD) {
	
		for (j = 16; j < 34; j++) {
			lcd_setpixel(40, j);
			lcd_setpixel(41, j);
		}
		
		for (j = 10; j < 30; j++) {
			lcd_setpixel(14, j);
			lcd_setpixel(15, j);
		}
		
		for (j = 50; j < 75; j++) {
			lcd_setpixel(j, 10);
			lcd_setpixel(j, 11);
		}
	}
}

/*Seteaza pozitia initiala a sarpelui si a punctului de mancare.*/
void snake_init(void) {

	int i;
	
	lungime = INIT_LEN;
	directia = DRT;
	
	snake[0].x = INIT_X;
	snake[0].y = INIT_Y;
	
	for (i = 1; i < lungime; i++) {
		snake[i].x = snake[0].x - i * SIZE;
		snake[i].y = snake[0].y;
	}
	
	food.x = rand() % (LCD_X_RES - (2 * SIZE)) + SIZE;
	food.y = rand() % (LCD_Y_RES - (2 * SIZE)) + SIZE;
	
	//	evita aparitia punctului de mancare in pereti pentru HARD
	if (play_mode == HARD) {
	
		char ok = 1;
		do {
			
			if (((food.x == 40) || (food.x == 40)) 
				&& (food.y < 34) && (food.y >= 16)) {
				ok = 0;	
			}
			
			if (((food.x == 14) || (food.x == 15)) 
				&& (food.y < 30) && (food.y >= 10)) {
				ok = 0;
			}
			
			if (((food.y == 10) || (food.y == 15)) 
				&& (food.x < 75) && (food.x >= 50)) {
				ok = 0;
			}
			
			if (!ok) { // regenereaza mancare intr-un punct random
				food.x = rand() % (LCD_X_RES - (2 * SIZE)) + SIZE;
				food.y = rand() % (LCD_Y_RES - (2 * SIZE)) + SIZE;
			}	
		} while (!ok);
	}
}

/* Pozitia sarpelui in urma deplasarii */
void snake_depl(char dir) {

	int i;
	
	/* Toata blocurile din coada urmaesc coordonatele blocurilor din fata lor. */
	for (i = lungime - 1; i > 0; i--) {
		snake[i].x = snake[i - 1].x; 
		snake[i].y = snake[i - 1].y;
	}
	
	/* Capul sarpelui urmareste directia data de jucator. */
	if (dir == SUS) {
	
		if (snake[0].y > 0) {
			snake[0].y -= SIZE;
		} else {
			snake[0].y = LCD_Y_RES - SIZE;
		}
		
	} else if (dir == JOS) {
	
		if (snake[0].y < LCD_Y_RES - SIZE) {
			snake[0].y += SIZE;
		} else {
			snake[0].y = 0;
		}
		
	} else if (dir == STG) {
	
		if (snake[0].x > 0) {
			snake[0].x -= SIZE;
		} else {
			snake[0].x = LCD_X_RES - SIZE;
		}
	
	} else if (dir == DRT) {
	
		if (snake[0].x < LCD_X_RES - SIZE) {
			snake[0].x += SIZE;
		} else {
			snake[0].x = 0;
		}
	}
}

void my_delay(void) {
	volatile int i;
	for (i = -32000; i < 32000; i++);
}
void SnakeDelay(void) {
	volatile int i;
	for (i = -32000; i < 32000; i++);
}

void SnakeDelay_mod_3(void) {
	volatile int i;
	int scd = scor * 1000;	
	for (i = -20000; i < (20000 - scd); i++);
}

/* Functie ce afiseaza pe lcd meniul */
void print_selected_level(void) {

	lcd_clear();
	lcd_update();
	
	lcd_gotoXY(2, 0);
	lcd_str("Select level", 12);
	
	lcd_gotoXY(10, 2);
	if (play_mode == EASY)
		lcd_str(">> Easy", 7);
	else
		lcd_str("   Easy", 7);
	
	lcd_gotoXY(10, 3);
	if (play_mode == MEDIUM)
		lcd_str(">> Medium", 9);
	else
		lcd_str("   Medium", 9);

	lcd_gotoXY(10, 4);
	if (play_mode == HARD)
		lcd_str(">> Hard", 7);
	else
		lcd_str("   Hard", 7);
}

/* Functie prin care se alege modul de joc */
void select_mode(void) {
	print_selected_level();

	while(1) {
	
		// s-a selectat mod de joc
		if ((!(PINB & (1 << PB0)))) {
			SnakeDelay(); 
			break;
		}
		
		// jos
		if ((!(PINC & (1 << PC0)))) {
			play_mode = (play_mode + 1) % 3;
			print_selected_level();
		}
		
		// sus
		if ((!(PINC & (1 << PC1)))) {
			play_mode = (play_mode - 1);
			if(play_mode == -1)
				play_mode = 2;
			print_selected_level();
		}
		
		SnakeDelay(); 
		
	}
	scor = 0;
	SnakeDelay();
	lcd_clear();
	lcd_update();
	snake_init();
	
}

/*Afiseaza Game Over + Scor la termiarea jocului.*/
void game_over(void) {

	char buf[2];
	
	itoa(scor, buf, 10);
	
	lcd_init();
	lcd_clear();
	lcd_update();
	
	lcd_gotoXY(11, 1);	
	lcd_str("Game Over", 9);
	
	lcd_gotoXY(15, 3);	
	lcd_str("SCOR:", 5);
	
	lcd_gotoXY(50, 3);
	if (scor < 10)
		lcd_str(buf, 1);
	else 
		lcd_str(buf, 2);
		
	while(1){
	
		// reset
		if (!(PINB & (1 << PB0))) {
			my_delay();
			select_mode();
			break;
		}
		
	}
	
}

/*
	Verifica daca nu a murit sarpele sau daca a ajuns la un punct de mancare.
*/

void check_snake(void) {

	int i;
		
	// sarpele a iesit din cadran (pentru MEDIUM)
	if ((play_mode == MEDIUM) || (play_mode == HARD)) {
		if ((snake[0].x <= 0) || (snake[0].x >= LCD_X_RES)
			|| (snake[0].y <= 0) || (snake[0].y >= LCD_Y_RES)) {		
				game_over();
		}
	}
	
	// sarpele s-a lovit in zidul de pe harta
	if (play_mode == HARD) {
	
		if (((snake[0].x == 40) || (snake[0].x == 40)) 
			&& (snake[0].y < 34) && (snake[0].y >= 16)) {
				game_over();
		}

		if (((snake[0].x == 14) || (snake[0].x == 15)) 
			&& (snake[0].y < 30) && (snake[0].y >= 10)) {
				game_over();
		}
		
		if (((snake[0].y == 10) || (snake[0].y == 15)) 
			&& (snake[0].x < 75) && (snake[0].x >= 50)) {
				game_over();
		}

	}
	
	// sarpele "s-a muscat" pe el
	for (i = 1; i < lungime; i++) {
		if ((snake[0].x == snake[i].x) && (snake[0].y == snake[i].y)) {
			game_over();	
		}
	}
	
	/* Sarpele a mancat mancarea */
	if (((snake[0].x <= food.x + SIZE) && (snake[0].y <= food.y + SIZE) 
		&& (snake[0].x >= food.x) && (snake[0].y >= food.y)) 
		|| ((snake[0].x + SIZE <= food.x + SIZE) && (snake[0].y + SIZE <= food.y + SIZE) 
		&& (snake[0].x + SIZE >= food.x) && (snake[0].y + SIZE >= food.y))) {
	
		// mareste dimensiunea sarpelui
		snake[lungime].x = snake[lungime - 1].x;
		snake[lungime].y = snake[lungime - 1].y;
		lungime++;

		// regenereaza mancare intr-un punct random
		food.x = rand() % (LCD_X_RES - (2 * SIZE)) + SIZE;
		food.y = rand() % (LCD_Y_RES - (2 * SIZE)) + SIZE;
		
		// evita aparitia punctului de mancare in pereti pentru HARD
		if (play_mode == HARD) {
		
			char ok = 1;
		
			do {
					
				if (((food.x == 40) || (food.x == 40)) 
					&& (food.y < 34) && (food.y >= 16)) {
					ok = 0;		
				}

				if (((food.x == 14) || (food.x == 15)) 
					&& (food.y < 30) && (food.y >= 10)) {
					ok = 0;
				}
					
				if (((food.y == 10) || (food.y == 15)) 
					&& (food.x < 75) && (food.x >= 50)) {
					ok = 0;
				}
				
				if (!ok) { // regenereaza mancare intr-un punct random
					food.x = rand() % (LCD_X_RES - (2 * SIZE)) + SIZE;
					food.y = rand() % (LCD_Y_RES - (2 * SIZE)) + SIZE;
				}
				
			} while (!ok);
		
		}
		
		// update scor
		scor++;
		
	}
	
}


void check_direction(void) {
		
	// sus
	if (!(PINC & (1 << PC1))) {
		newDirection = SUS;
	}
		
	// jos
	if (!(PINC & (1 << PC0))) {
		newDirection = JOS;
	}
		
	// stanga
	if (!(PIND & (1 << PD1))) {
		newDirection = STG;
	}
		
	// dreapta
	if (!(PIND & (1 << PD0))) {
		newDirection = DRT;
	}
	
}


int main(void) {
	
	lcd_init();
	lcd_clear();
	lcd_update();
	
	DDRC &= ~(1 << PC0); // jos
	PORTC |= (1 << PC0);
	
	DDRD &= ~(1 << PD0); // dreapta
	PORTD |= (1 << PD0);
	
	DDRC &= ~(1 << PC1); // sus
	PORTC |= (1 << PC1);
	
	DDRD &= ~(1 << PD1); // stanga
	PORTD |= (1 << PD1);
	
	DDRB &= ~(1 << PB0); // select
	PORTB |= (1 << PB0);
	
	timer_init();
	/* Selecteaza dificultatea jocului */
	select_mode();

	while(1) {
	
		lcd_clear();
		newDirection = directia;
		
		check_direction();
		
		// reset
		if ((!(PINB & (1 << PB0)))) {
			my_delay();
			select_mode();
			continue;
		}
		
		if ((newDirection == SUS) && (directia != JOS))
			directia = newDirection;
			
		if ((newDirection == JOS) && (directia != SUS)) 
			directia = newDirection;
			
		if ((newDirection == STG) && (directia != DRT)) 
			directia = newDirection;
			
		if ((newDirection == DRT) && (directia != STG)) 
			directia = newDirection;
			
		snake_depl(directia);
		check_snake();
		
		drawBlocks();
		
		if (play_mode == HARD)
			SnakeDelay_mod_3();
		else
			SnakeDelay();
			
		lcd_update();
		
	}

}